# Error Pages

As in any app, errors may well happen in Perseus, and you should be prepared for them. However, because we're building a user-facing app, we need to be able to display pretty error pages to the user to explain what went wrong and what they can do to fix it (if applicable).

In Perseus, error pages are defined by HTTP status codes, and you can define others as necessary for your own internal errors, Perseus is only concerned with those that may occur in communication with the backend.

## Defining Error Pages

The easiest way of defining your error pages is to put them into a function like `get_error_pages`, like so:

```rust
use sycamore::prelude::template;
use perseus::ErrorPages;

pub fn get_error_pages() -> ErrorPages {
    let mut error_pages = ErrorPages::new(Box::new(|_, _, _| {
        template! {
            p { "Another error occurred." }
        }
    }));
    error_pages.add_page(
        404,
        Box::new(|_, _, _| {
            template! {
                p { "Page not found." }
            }
        }),
    );
    error_pages.add_page(
        400,
        Box::new(|_, _, _| {
            template! {
                p { "Client error occurred..." }
            }
        }),
    );

    error_pages
}
```

This example creates three error pages, the first of which is mandatory: a fallback page. There are [a lot](https://httpstatuses.com) of HTTP status codes, and it would be ridiculous for you to handle them all. However, occasionally you may get one that you didn't anticipate, and you should be ready for it.

The rest of the error pages are defined specifically for error codes, like `404` and `400` above.

Every error page is a function that takes three arguments: the URL that caused the problem, the HTTP status code, and the error message that was the payload of the request.

## Best Practices

When designing error pages, remember that they should be intelligible to everyone, regardless of technical ability! To that end, the error message provided by the serve should probably only be accessible through a dropdown or the like that provides technical details so the user can report it.

## Security Concerns

Error messages can leak sensitive data, so Perseus tries to make sure that they don't! No Perseus error message will ever provide details more than the rendering strategy a page uses (which shouldn't be sensitive...), however they will include error messages generated by your custom rendering functions. Essentially, assume any error message generated and returned in the rendering process may at some point be visible to the client, so **don't put sensitive information in your error messages**!
